from .polly_synthesizer import PollySynthesizer
from .elevenlabs_synthesizer import ElevenlabsSynthesizer
from .openai_synthesizer import OPENAISynthesizer
from .deepgram_synthesizer import DeepgramSynthesizer
from .azure_synthesizer import AzureSynthesizer
from .cartesia_synthesizer import CartesiaSynthesizer
from .smallest_synthesizer import SmallestSynthesizer
